"""Project metadata.
"""

__title__ = 'alchy'
__summary__ = ('A SQLAlchemy extension for its declarative ORM that provides '
               'enhancements for model classes, queries, and sessions.')
__url__ = 'https://github.com/dgilland/alchy'

__version__ = '2.2.1'
__install_requires__ = ['SQLAlchemy>=0.9.0', 'pydash>=3.4.3']

__author__ = 'Derrick Gilland'
__email__ = 'dgilland@gmail.com'

__license__ = 'MIT License'
